mod var;

use std::ops::{Deref, DerefMut};

use bevy_ecs::{
    entity::Entity,
    prelude::{Bundle, Component},
    query::{QueryEntityError, With},
    system::{Commands, Query, SystemParam},
};

pub use self::var::*;

pub type Scalar = f64;

#[derive(Default, Clone, Debug, Component)]
pub struct Point<V = VarId> {
    pub x: V,
    pub y: V,
}

impl<V> Point<V> {
    pub fn new(x: V, y: V) -> Self {
        Self { x, y }
    }
}

impl<V> From<(V, V)> for Point<V> {
    fn from(p: (V, V)) -> Self {
        Self { x: p.0, y: p.1 }
    }
}

pub type PointPos = Point<Scalar>;
pub type PointId = Entity;

#[derive(Clone, Debug, Default, Component)]
// TODO: segment, ray and full line
// TODO: what if start and end are coincident?
pub struct Line<P = PointId> {
    pub start: P,
    pub end: P,
}

impl<P> Line<P> {
    pub fn new(start: P, end: P) -> Self {
        Self { start, end }
    }
}

pub type LinePos = Line<PointPos>;

#[derive(Clone, Component)]
// TODO: arc. how to represent start and end of arc?
// angles, unit vectors, or a point for each?
pub struct Circle {
    pub center: PointId,
    pub radius: VarId,
}

// pub type CirclePos = CircleEntity<PointPos>;

#[derive(Bundle)]
pub struct PointBundle {
    pub point: Point,
    pub computed_pos: ComputedPointPos,
}

impl PointBundle {
    pub fn new(x: VarId, y: VarId) -> Self {
        Self {
            point: Point::new(x, y),
            computed_pos: Default::default(),
        }
    }
}

pub fn insert_point_at(commands: &mut Commands, point: impl Into<PointPos>) -> PointId {
    let point = point.into();
    let x = commands.spawn(Var::new_free(point.x)).id();
    let y = commands.spawn(Var::new_free(point.y)).id();
    commands.spawn(PointBundle::new(x, y)).id()
}

#[derive(Component, Default)]
pub struct ComputedPointPos(Var<PointPos>);

impl Deref for ComputedPointPos {
    type Target = Var<PointPos>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

pub fn update_point_pos(mut points: Query<(&Point, &mut ComputedPointPos)>, vars: Query<&Var>) {
    points.for_each_mut(|(point, mut computed_pos)| {
        let x = vars.get(point.x).unwrap();
        let y = vars.get(point.y).unwrap();
        computed_pos.0 = x.merge(*y, PointPos::new);
    });
}

#[derive(SystemParam)]
pub struct PointPosQueryMut<'w, 's> {
    vars: Query<'w, 's, &'static mut Var>,
}

impl<'w, 's> PointPosQueryMut<'w, 's>
// impl<'w, 's, R> PointPosQueryImpl<'w, 's, Query<'w, 's, R>>
// where
//     R: WorldQuery,
//     for<'w2, 's2> <<R as WorldQuery>::Fetch as Fetch<'w2, 's2>>::Item: DerefMut<Target = Var>,
{
    pub fn get(&self, point: &Point) -> Result<Var<PointPos>, QueryEntityError> {
        let x = self.vars.get(point.x)?;
        let y = self.vars.get(point.y)?;
        Ok(x.merge(*y, PointPos::new))
    }

    pub fn set(&mut self, point: &Point, move_to: PointPos) -> Result<(), QueryEntityError> {
        for (id, val) in [(point.x, move_to.x), (point.y, move_to.y)] {
            self.vars.get_mut(id)?.deref_mut().value = val;
        }
        Ok(())
    }
}

#[derive(Default, Component)]
pub struct ComputedLinePos(Var<LinePos>);

impl Deref for ComputedLinePos {
    type Target = Var<LinePos>;

    fn deref(&self) -> &Self::Target {
        &self.0
    }
}

pub fn update_line_pos(
    mut lines: Query<(&Line, &mut ComputedLinePos)>,
    points: Query<&ComputedPointPos>,
) {
    lines.for_each_mut(|(line, mut computed_pos)| {
        let Ok(start) = points.get(line.start) else { return };
        let Ok(end) = points.get(line.end) else { return} ;
        computed_pos.0 = (*start).clone().merge((*end).clone(), LinePos::new);
    });
}

#[derive(Bundle)]
pub struct LineBundle {
    pub line: Line,
    pub computed_pos: ComputedLinePos,
}

impl LineBundle {
    pub fn new(start: PointId, end: PointId) -> Self {
        Self {
            line: Line::new(start, end),
            computed_pos: Default::default(),
        }
    }
}

pub fn remove_dangling_lines(
    mut lines: Query<(Entity, &Line)>,
    points: Query<(), With<Point>>,
    mut commands: Commands,
) {
    lines.for_each_mut(|(id, line)| {
        if !points.contains(line.start) || !points.contains(line.end) {
            commands.entity(id).despawn();
        }
    });
}
