#![feature(generic_const_exprs)]

use bevy_ecs::prelude::*;
use eframe::egui;
use rand::seq::SliceRandom;
use rand_distr::Distribution;

use sketchrs::geometry;

#[cfg(feature = "trace")]
mod trace;
mod ui;

fn init(mut commands: Commands) {
    // let p1 = geometry::insert_point_at(&mut commands, (10., 30.));
    // let p2 = geometry::insert_point_at(&mut commands, (-20., 15.));
    // geometry::insert_point_at(&mut commands, (0., -10.));
    // commands.spawn(geometry::LineBundle::new(p1, p2));

    // Insert 1000 random points
    let dist = rand_distr::Uniform::new(-100., 100.);
    let mut rng = rand::thread_rng();
    let mut points = Vec::new();
    for _ in 0..1000 {
        let p = geometry::insert_point_at(
            &mut commands,
            (dist.sample(&mut rng), dist.sample(&mut rng)),
        );
        points.push(p);
    }

    // Insert 1000 random lines
    for _ in 0..1000 {
        let p1 = *points.choose(&mut rng).unwrap();
        let p2 = *points.choose(&mut rng).unwrap();
        commands.spawn(geometry::LineBundle::new(p1, p2));
    }
}

struct MyApp {
    world: World,
    update_schedule: Schedule,
}

impl Default for MyApp {
    fn default() -> Self {
        Self::new()
    }
}

impl MyApp {
    fn new() -> Self {
        let mut world = World::default();

        #[cfg(feature = "trace")]
        trace::setup(&mut world);

        ui::init(&mut world);

        let mut init_sched = Schedule::new();
        init_sched.add_system(init);
        init_sched.run(&mut world);

        let mut update_schedule = Schedule::default();
        ui::add_to_schedule(&mut update_schedule);

        Self {
            world,
            update_schedule,
        }
    }
}

impl eframe::App for MyApp {
    fn update(&mut self, ctx: &egui::Context, _frame: &mut eframe::Frame) {
        let span = tracing::info_span!("update");
        let _enter = span.enter();

        self.world.insert_resource(ui::ContextRes(ctx.clone()));

        self.update_schedule.run(&mut self.world);

        self.world.remove_resource::<ui::ContextRes>();
    }
}

fn main() {
    tracing_subscriber::fmt()
        .with_max_level(tracing::Level::INFO)
        .with_target(false)
        .init();

    let options = eframe::NativeOptions::default();
    eframe::run_native("sketchrs", options, Box::new(|_cc| Box::<MyApp>::default())).unwrap();
}
