use std::{borrow::Borrow, fmt::Debug, ops::Index};

mod params;
mod residuals;

use params::ReadParams;

use argmin_math;
use levenberg_marquardt::{self, TerminationReason};
use nalgebra::{ArrayStorage, Const, Dyn, VecStorage};
use nalgebra_sparse::CooMatrix;

use crate::geometry::Scalar;

pub type Idx = u32;

use self::{params::{Parameters, Param}, residuals::WriteResiduals};

type SVector<const D: usize> = nalgebra::SVector<Scalar, D>;
type SMatrix<const R: usize, const C: usize> = nalgebra::SMatrix<Scalar, R, C>;
type DVector = nalgebra::DVector<Scalar>;
type DMatrix = nalgebra::DMatrix<Scalar>;


trait CostFunction {
    type Params;
    // const NUM_RESIDUALS: usize;
    type Residuals;
    type Jacobian;

    fn evaluate(
        &self,
        params: &Self::Params,
        residual: Option<&mut Self::Residuals>,
        jacobian: Option<&mut Self::Jacobian>,
    );

    fn residual(&self, x: &Self::Params) -> Self::Residuals
    where
        Self::Residuals: Default,
    {
        let mut residual = Self::Residuals::default();
        self.evaluate(x, Some(&mut residual), None);
        residual
    }

    fn jacobian(&self, x: &Self::Params) -> Self::Jacobian
    where
        Self::Jacobian: Default,
    {
        let mut jacobian = Self::Jacobian::default();
        self.evaluate(x, None, Some(&mut jacobian));
        jacobian
    }

    fn with_indices<PI, RI, JI>(
        self,
        param_indices: PI,
        residual_indices: RI,
        jacobian_indices: JI,
    ) -> IndexedCostFunctionImpl<Self, PI, RI, JI>
    where
        Self: Sized,
    {
        IndexedCostFunctionImpl {
            residual: self,
            param_indices,
            residual_indices,
            jacobian_indices,
        }
    }
}

trait IndexedCostFunction {
    fn evaluate(
        &self,
        params: &Parameters,
        residuals: Option<&mut [Scalar]>,
        jacobian: Option<&mut [Scalar]>,
    );
}

struct IndexedCostFunctionImpl<R, PI, RI, JI> {
    residual: R,
    param_indices: PI,
    residual_indices: RI,
    jacobian_indices: JI,
}

impl<R, PI, RI, JI> IndexedCostFunction for IndexedCostFunctionImpl<R, PI, RI, JI>
where
    R: CostFunction,
    R::Params: ReadParams<PI>,
    R::Residuals: WriteResiduals<RI> + Default,
    R::Jacobian: WriteResiduals<JI> + Default,
{
    fn evaluate(
        &self,
        params: &Parameters,
        residuals: Option<&mut [Scalar]>,
        jacobian: Option<&mut [Scalar]>,
    ) {
        // let p = R::Params::read(params, self.param_indices);

        // let mut res = residuals.and_then(|_| Some(R::Residuals::default()));
        // let mut jac = jacobian.and_then(|_| Some(R::Jacobian::default()));
        // self.residual.evaluate(&p, res.as_mut(), jac.as_mut());

        // if let (Some(res), Some(residuals)) = (res, residuals) {
        //     res.write(self.residual_indices, residuals);
        // }
        // if let (Some(jac), Some(jacobian)) = (jac, jacobian) {
        //     jac.write(self.jacobian_indices, jacobian);
        // }
    }
}

slotmap::new_key_type!{ struct ParamKey; }

struct ParamIndex(Idx);

#[derive(Default)]
struct CostFunctions {
    functions: Vec<Box<dyn IndexedCostFunction>>,
    param_indices: slotmap::SlotMap<ParamKey, Idx>,
    num_residuals: Idx,
}

impl CostFunctions {
    fn new() -> Self {
        Self::default()
    }

    // fn add_param<P: Param>(&mut self) -> ParamIndex {
    //     // let index = self.params_size;
    //     // self.params_size += P::SIZE;
    //     ParamIndex(index)
    // }

    // fn add<R, PI, RI, JI>(&mut self, residual: R, param_indices: PI, residual_indices: RI, jacobian_indices: JI)
    // where
    //     R: CostFunction,
    //     R::Params: ReadParams<PI>,
    //     R::Residuals: WriteResiduals<RI> + Default,
    //     R::Jacobian: WriteResiduals<JI> + Default,
    // {
    //     self.functions.push(Box::new(
    //         residual
    //             .with_indices(param_indices, residual_indices, jacobian_indices),
    //     ));
    //     self.param_indices.push(param_indices.into());
    //     self.residual_indices.push(residual_indices.into());
    //     self.jacobian_indices.push(jacobian_indices.into());
    // }

    fn evaluate(&self, params: &Parameters, residuals: Option<&mut [Scalar]>, jacobian: Option<&mut [Scalar]>) {
        // for (i, f) in self.functions.iter().enumerate() {
        //     f.evaluate(params, residuals, jacobian);
        // }
    }
}

trait SResBlock<const NIN: usize, const NOUT: usize = 1>: Clone + Debug {
    fn evaluate(
        &self,
        x: SVector<NIN>,
        residual: Option<&mut SVector<NOUT>>,
        jacobian: Option<&mut SMatrix<NIN, NOUT>>,
    ) {
        if let Some(residual) = residual {
            *residual = self.residual(x);
        }
        if let Some(jacobian) = jacobian {
            *jacobian = self.jacobian(x);
        }
    }

    fn residual(&self, x: SVector<NIN>) -> SVector<NOUT> {
        let mut residual = SVector::<NOUT>::zeros();
        self.evaluate(x, Some(&mut residual), None);
        residual
    }
    fn jacobian(&self, x: SVector<NIN>) -> SMatrix<NIN, NOUT> {
        let mut jacobian = SMatrix::<NIN, NOUT>::zeros();
        self.evaluate(x, None, Some(&mut jacobian));
        jacobian
    }

    // Probably should move to own trait
    // fn into_indexed(
    //     self,
    //     in_indices: [Index; NIN],
    //     out_indices: [Index; NOUT],
    // ) -> Box<dyn IndexedResBlock>
    // where
    //     Self: Sized + 'static,
    // {
    //     Box::new(IndexedResBlockImpl {
    //         residual: self,
    //         in_indices,
    //         out_indices,
    //     })
    // }
}

fn square(x: Scalar) -> Scalar {
    x * x
}

#[derive(Clone, Debug)]
pub struct Equal1D;

impl SResBlock<2> for Equal1D {
    fn evaluate(
        &self,
        x: SVector<2>,
        residual: Option<&mut SVector<1>>,
        jacobian: Option<&mut SMatrix<2, 1>>,
    ) {
        let [x1, x2] = *x.as_ref();
        let dx = x2 - x1;
        if let Some(residual) = residual {
            residual[0] = square(dx);
        }
        if let Some(jacobian) = jacobian {
            jacobian[0] = -2. * dx;
            jacobian[1] = 2. * dx;
        }
    }
}

#[derive(Clone, Debug)]
pub struct AxisDistance(Scalar);

/*
impl ResBlock for AxisDistance {
    fn parameter_sizes(&self) -> &[usize] {
        &[2]
    }

    fn n_residuals(&self) -> usize {
        1
    }

    fn evaluate(
        &self,
        x: &[&[Scalar]],
        residual: Option<&mut [Scalar]>,
        jacobian: Option<&mut [&mut [Scalar]]>,
    ) {
        let [x1, x2] = x[0] else { panic!() };
        let dx = x2 - x1;
        let r = self.0;
        if let Some(residual) = residual {
            residual[0] = square(dx) - square(r);
        }
        if let Some(jacobian) = jacobian {
            jacobian[0][0] = -2. * dx;
            jacobian[0][1] = 2. * dx;
        }
    }
}
*/

impl SResBlock<2> for AxisDistance {
    fn evaluate(
        &self,
        x: SVector<2>,
        residual: Option<&mut SVector<1>>,
        jacobian: Option<&mut SMatrix<2, 1>>,
    ) {
        let [x1, x2] = *x.as_ref();
        let dx = x2 - x1;
        let r = self.0;
        if let Some(residual) = residual {
            residual[0] = square(dx) - square(r);
        }
        if let Some(jacobian) = jacobian {
            jacobian[0] = -2. * dx;
            jacobian[1] = 2. * dx;
        }
    }
}

#[derive(Clone, Debug)]
pub struct PointPointDistance(Scalar);

impl CostFunction for PointPointDistance {
    type Params = (SVector<2>, SVector<2>);
    type Residuals = Scalar;
    type Jacobian = (SMatrix<2, 1>, SMatrix<2, 1>);

    fn evaluate(
        &self,
        (p1, p2): &Self::Params,
        residual: Option<&mut Self::Residuals>,
        jacobian: Option<&mut Self::Jacobian>,
    ) {
        let d = p2 - p1;
        if let Some(residual) = residual {
            *residual = square(d.x) + square(d.y) - square(self.0);
        }
        if let Some(jacobian) = jacobian {
            jacobian.0.copy_from(&(2. * d));
            jacobian.1.copy_from(&(-2. * d));
            // jacobian.row_mut(0).copy_from(&(2. * d).transpose());
            // jacobian.row_mut(1).copy_from(&(-2. * d).transpose());
            // jacobian[0] = -2. * d.x;
            // jacobian[1] = -2. * d.y;
            // jacobian[2] = 2. * d.x;
            // jacobian[3] = 2. * d.y;
        }
    }
}

impl SResBlock<4> for PointPointDistance {
    fn evaluate(
        &self,
        x: SVector<4>,
        residual: Option<&mut SVector<1>>,
        jacobian: Option<&mut SMatrix<4, 1>>,
    ) {
        let [x1, y1, x2, y2] = *x.as_ref();
        let (dx, dy) = (x2 - x1, y2 - y1);
        if let Some(residual) = residual {
            residual[0] = square(dx) + square(dy) - square(self.0);
        }
        if let Some(jacobian) = jacobian {
            jacobian[0] = -2. * dx;
            jacobian[1] = -2. * dy;
            jacobian[2] = 2. * dx;
            jacobian[3] = 2. * dy;
        }
    }
}

#[derive(Clone, Debug)]
pub struct PointDistance(SVector<2>, Scalar);

impl SResBlock<2> for PointDistance {
    fn evaluate(
        &self,
        x: SVector<2>,
        residual: Option<&mut SVector<1>>,
        jacobian: Option<&mut SMatrix<2, 1>>,
    ) {
        let d = x - self.0;
        let r = self.1;
        if let Some(residual) = residual {
            residual[0] = d.norm_squared() - square(r);
        }
        if let Some(jacobian) = jacobian {
            jacobian.copy_from(&(2. * d));
        }
    }
}

#[derive(Clone, Debug)]
pub struct Collinear;

impl CostFunction for Collinear {
    type Params = (SVector<2>, SVector<2>, SVector<2>);
    type Residuals = Scalar;
    type Jacobian = (SVector<2>, SVector<2>, SVector<2>);

    fn evaluate(
        &self,
        (p1, p2, p3): &Self::Params,
        residual: Option<&mut Self::Residuals>,
        jacobian: Option<&mut Self::Jacobian>,
    ) {
        let d21 = p2 - p1;
        let d32 = p3 - p2;
        if let Some(residual) = residual {
            *residual = d21.y * d32.x - d32.y * d21.x;
        }
        if let Some(jacobian) = jacobian {
            jacobian.0.x = d32.y;
            jacobian.0.y = -d32.x;
            jacobian.1.x = p1.y - p3.y;
            jacobian.1.y = p3.x - p1.x;
            jacobian.2.x = d21.y;
            jacobian.2.y = -d21.x;
        }
    }
}

impl SResBlock<6> for Collinear {
    fn evaluate(
        &self,
        x: SVector<6>,
        residual: Option<&mut SVector<1>>,
        jacobian: Option<&mut SMatrix<6, 1>>,
    ) {
        let [x1, y1, x2, y2, x3, y3] = *x.as_ref();
        let (dx21, dy21) = (x2 - x1, y2 - y1);
        let (dx32, dy32) = (x3 - x2, y3 - y2);
        if let Some(residual) = residual {
            // dy21 / dx21 = dy32 / dx32
            // dy21 * dx32 = dy32 * dx21
            residual[0] = dy21 * dx32 - dy32 * dx21;
        }
        if let Some(jacobian) = jacobian {
            jacobian[0] = dy32;
            jacobian[1] = -dx32;
            jacobian[2] = y1 - y3;
            jacobian[3] = x3 - x1;
            jacobian[4] = dy21;
            jacobian[5] = -dx21;
        }
    }
}


trait ScatterResBlock: Debug {
    fn shape(&self) -> (usize, usize);
    fn apply_into(&self, x: &[Scalar], residuals: &mut [Scalar]);
    fn jacobian_into(&self, x: &[Scalar], jacobian: &mut [Scalar]);
    fn evaluate_into(&self, x: &[Scalar], residuals: &mut [Scalar], jacobian: &mut [Scalar]) {
        self.apply_into(x, residuals);
        self.jacobian_into(x, jacobian);
    }
}

#[derive(Clone, Debug)]
struct ScatterResBlockImpl<const NIN: usize, const NOUT: usize, R> {
    residual: R,
    in_indices: [Idx; NIN],
    residual_indices: [Idx; NOUT],
    jacobian_indices: [[Idx; NIN]; NOUT],
}

impl<const NIN: usize, const NOUT: usize, R> ScatterResBlockImpl<NIN, NOUT, R> {
    fn gather_svector(&self, x: &[Scalar]) -> SVector<NIN> {
        SVector::from_fn(|i, _| x[self.in_indices[i] as usize])
    }
}

impl<const NIN: usize, const NOUT: usize, R> ScatterResBlock for ScatterResBlockImpl<NIN, NOUT, R>
where
    R: SResBlock<NIN, NOUT>,
{
    fn shape(&self) -> (usize, usize) {
        (NIN, NOUT)
    }

    fn apply_into(&self, x: &[Scalar], residuals: &mut [Scalar]) {
        // Front load bounds check
        let mut inbounds = true;
        for i in self.residual_indices.iter().cloned() {
            inbounds &= (i as usize) < residuals.len();
        }
        assert!(inbounds);

        let res = self.residual.residual(self.gather_svector(x));
        for (i, r) in self
            .residual_indices
            .iter()
            .cloned()
            .zip(res.borrow().iter())
        {
            residuals[i as usize] = *r;
        }
    }

    fn jacobian_into(&self, x: &[Scalar], jacobian: &mut [Scalar]) {
        // Front load bounds check
        let mut inbounds = true;
        for ir in self.jacobian_indices.iter() {
            for i in ir.iter().cloned() {
                inbounds &= (i as usize) < jacobian.len();
            }
        }
        assert!(inbounds);

        let jac = self.residual.jacobian(self.gather_svector(x));
        for (ir, r) in self.jacobian_indices.iter().zip(jac.column_iter()) {
            for (i, r) in ir.iter().zip(r.iter()) {
                jacobian[*i as usize] = *r;
            }
        }
    }

    fn evaluate_into(&self, x: &[Scalar], residuals: &mut [Scalar], jacobian: &mut [Scalar]) {
        // Front load bounds check
        let mut inbounds = true;
        for i in self.residual_indices.iter().cloned() {
            inbounds &= (i as usize) < residuals.len();
        }
        for ir in self.jacobian_indices.iter() {
            for i in ir.iter().cloned() {
                inbounds &= (i as usize) < jacobian.len();
            }
        }
        assert!(inbounds);

        let mut res = SVector::<NOUT>::zeros();
        let mut jac = SMatrix::<NIN, NOUT>::zeros();
        self.residual
            .evaluate(self.gather_svector(x), Some(&mut res), Some(&mut jac));
        for (i, r) in self.residual_indices.iter().cloned().zip(res.iter()) {
            residuals[i as usize] = *r;
        }
        for (ir, r) in self.jacobian_indices.iter().zip(jac.column_iter()) {
            for (i, r) in ir.iter().zip(r.iter()) {
                jacobian[*i as usize] = *r;
            }
        }
    }
}

/*
trait IndexedResBlock {
    fn shape(&self) -> (usize, usize);
    fn in_indices(&self) -> &[Index];
    fn out_indices(&self) -> &[Index];
    fn into_scatter(self: Box<Self>, total_shape: (usize, usize)) -> Box<dyn ScatterResBlock>;
}

struct IndexedResBlockImpl<const NIN: usize, const NOUT: usize, R> {
    residual: R,
    in_indices: [Index; NIN],
    out_indices: [Index; NOUT],
}

impl<const NIN: usize, const NOUT: usize, R> IndexedResBlock for IndexedResBlockImpl<NIN, NOUT, R>
where
    R: SResBlock<NIN, NOUT> + 'static,
{
    fn shape(&self) -> (usize, usize) {
        (NIN, NOUT)
    }

    fn in_indices(&self) -> &[Index] {
        &self.in_indices
    }

    fn out_indices(&self) -> &[Index] {
        &self.out_indices
    }

    fn into_scatter(self: Box<Self>, total_shape: (usize, usize)) -> Box<dyn ScatterResBlock> {
        let mut jacobian_indices = [[0; NIN]; NOUT];
        for (i, jr) in jacobian_indices.iter_mut().enumerate() {
            for (j, jj) in jr.iter_mut().enumerate() {
                // *jj = self.in_indices[j] + (self.out_indices[i] * total_shape.0 as u32) as Index;
                *jj = self.out_indices[i] + (self.in_indices[j] * total_shape.0 as u32) as Index;
            }
        }
        Box::new(ScatterResBlockImpl {
            residual: self.residual,
            in_indices: self.in_indices,
            residual_indices: self.out_indices,
            jacobian_indices,
        })
    }
}
*/

#[derive(Debug)]
struct ResidualBlocks {
    res_blocks: Vec<Box<dyn ScatterResBlock>>,
    shape: (usize, usize),
}

impl ResidualBlocks {
    fn new(res_blocks: Vec<Box<dyn IndexedCostFunction>>, shape: (usize, usize)) -> Self {
        // let res_blocks = res_blocks
        //     .into_iter()
        //     .map(|r| r.into_scatter(shape))
        //     .collect();
        // Self { res_blocks, shape }
        todo!()
    }

    fn residuals_into<X: AsRef<[Scalar]>>(&self, x: X, residuals: &mut [Scalar]) {
        let x = x.as_ref();
        assert!(self.shape.0 == residuals.len());
        for residual in &self.res_blocks {
            residual.apply_into(x, residuals);
        }
    }

    fn residuals<X: AsRef<[Scalar]>>(&self, x: X) -> Vec<Scalar> {
        let mut residuals = vec![0.; self.shape.0];
        self.residuals_into(x, &mut residuals);
        residuals
    }

    // fn jacobian_into<'a, X: AsRef<[Scalar]>, J: Into<DMatrixViewMut<'a>>>(
    //     &self,
    //     x: X,
    //     jacobian: J,
    // ) {
    //     let x = x.as_ref();
    //     let mut jacobian = jacobian.into();
    //     assert!(self.shape == jacobian.shape());
    //     for residual in &self.residuals {
    //         residual.jacobian_into(x, jacobian.as_mut_slice());
    //     }
    // }

    fn jacobian<X: AsRef<[Scalar]>>(&self, x: X) -> DMatrix {
        // assert!(self.shape == jacobian.shape());
        let x = x.as_ref();
        let mut jacobian = DMatrix::zeros(self.shape.0, self.shape.1);
        for residual in &self.res_blocks {
            residual.jacobian_into(x, jacobian.as_mut_slice());
        }
        jacobian
    }
}

#[derive(Debug)]
struct Problem {
    residuals: ResidualBlocks,
    x: DVector,
}

impl levenberg_marquardt::LeastSquaresProblem<Scalar, Dyn, Dyn> for Problem {
    type ResidualStorage = VecStorage<Scalar, Dyn, Const<1>>;
    type JacobianStorage = VecStorage<Scalar, Dyn, Dyn>;
    type ParameterStorage = VecStorage<Scalar, Dyn, Const<1>>;

    fn set_params(&mut self, x: &DVector) {
        self.x = x.clone();
    }

    fn params(&self) -> DVector {
        self.x.clone()
    }

    fn residuals(&self) -> Option<DVector> {
        let res = self.residuals.residuals(self.x.as_slice());
        Some(DVector::from(res))
    }

    fn jacobian(&self) -> Option<nalgebra::Matrix<Scalar, Dyn, Dyn, Self::JacobianStorage>> {
        let jac = self.residuals.jacobian(self.x.as_slice());
        Some(jac)
    }
}

pub fn test() {
    let lm = levenberg_marquardt::LevenbergMarquardt::<f64>::new();

    let nvar = 4;
    let nres = 3;
    // let residuals = CostFunction::new(
    //     vec![
    //         PointDistance(SVector::<2>::new(1., 0.), 0.75).into_indexed([0, 1], [0]),
    //         PointPointDistance(1.0).into_indexed([0, 1, 2, 3], [2]),
    //         PointDistance(SVector::<2>::new(0., 0.), 1.25).into_indexed([2, 3], [1]),
    //     ],
    //     (nres, nvar),
    // );

    let dist = rand_distr::Normal::new(0., 1.0).unwrap();
    let prob = Problem {
        residuals: todo!(),
        x: DVector::from_distribution_generic(
            Dyn(nvar),
            Const::<1>,
            &dist,
            &mut rand::thread_rng(),
        ),
    };

    println!("Problem: {:?}", prob);

    let (mut prob, rep) = lm.minimize(prob);
    println!("Minimized: {:?}", prob.x);
    println!("Report: {:?}", rep);

    // // perturb the initial guess
    // let dist = rand_distr::Normal::new(0., 1.0).unwrap();
    // prob.x += SVector::<2>::from_distribution(&dist, &mut rand::thread_rng());
    // println!("Problem: {:?}", prob);

    // let (mut prob, rep) = lm.minimize(prob);
    // println!("Minimized: {:?}", prob.x);
    // println!("Report: {:?}", rep);
}
