use nalgebra::{SMatrix, SVector};

use crate::geometry::Scalar;

use super::Idx;

pub struct Parameters {
    values: Vec<Scalar>,
}

pub trait Param {
    const SIZE: usize;
    type Jacobian<const N: usize>;

    fn from_array(array: [Scalar; Self::SIZE]) -> Self;

    fn from_slice(slice: &[Scalar]) -> Self
    where
        [Scalar; Self::SIZE]: Sized,
        Self: Sized,
    {
        let array =
            TryInto::<[Scalar; Self::SIZE]>::try_into(slice).expect("Slice has wrong length");
        Self::from_array(array)
    }

    fn copy_to_array(&self, array: &mut [Scalar; Self::SIZE]);

    fn copy_to_slice(&self, slice: &mut [Scalar])
    where
        [Scalar; Self::SIZE]: Sized,
    {
        let array =
            TryInto::<&mut [Scalar; Self::SIZE]>::try_into(slice).expect("Slice has wrong length");
        self.copy_to_array(array)
    }
}

trait WithJacobian<const N: usize> {
    type Jacobian<const M: usize>;
}

trait Jacobian {}

impl Param for Scalar {
    const SIZE: usize = 1;

    type Jacobian<const N: usize> = SVector<Scalar, N>;

    fn from_slice(slice: &[Scalar]) -> Self {
        let [x] = slice else {
            panic!("Slice has wrong length");
        };
        *x
    }

    fn copy_to_slice(&self, slice: &mut [Scalar]) {
        let [x] = slice else {
            panic!("Slice has wrong length");
        };
        *x = *self;
    }

    fn copy_to_array(&self, array: &mut [Scalar; Self::SIZE]) {
        array[0] = *self;
    }

    fn from_array(array: [Scalar; Self::SIZE]) -> Self {
        array[0]
    }
}

pub trait ReadParams<I> {
    fn read(params: &Parameters, index: I) -> Self;
}

// impl ReadParams<Idx> for Scalar {
//     fn read(params: &Parameters, index: Idx) -> Self {
//         params.values[index as usize]
//     }
// }

impl<const M: usize> Param for [Scalar; M] {
    const SIZE: usize = M;
    type Jacobian<const N: usize> = SMatrix<Scalar, M, N>;

    fn from_array(array: [Scalar; Self::SIZE]) -> Self {
        unsafe { std::mem::transmute_copy(&array) }
    }

    fn copy_to_array(&self, slice: &mut [Scalar; Self::SIZE]) {
        slice.copy_from_slice(self)
    }
}

impl<const M: usize> Param for SMatrix<Scalar, M, 1> {
    const SIZE: usize = M;
    type Jacobian<const N: usize> = SMatrix<Scalar, M, N>;

    fn from_slice(slice: &[Scalar]) -> Self {
        Self::from_column_slice(slice)
    }

    fn from_array(array: [Scalar; Self::SIZE]) -> Self {
        Self::from_slice(&array)
    }

    fn copy_to_slice(&self, slice: &mut [Scalar]) {
        slice.copy_from_slice(self.as_slice())
    }

    fn copy_to_array(&self, array: &mut [Scalar; Self::SIZE]) {
        array.copy_from_slice(self.as_slice())
    }
}

impl<P: Param> ReadParams<Idx> for P
where
    [Scalar; P::SIZE]: Sized,
{
    fn read(params: &Parameters, index: Idx) -> Self {
        Self::from_slice(&params.values[index as usize..][..Self::SIZE])
    }
}

// impl<const R: usize, const C: usize> ReadParams<Idx> for SMatrix<Scalar, R, C> {
//     fn read(params: &Parameters, index: Idx) -> Self {
//         SMatrix::from_column_slice(&params.values[index as usize..][..R * C])
//     }
// }

// TODO: Implement for tuples of arbitrary length using a macro
impl<T1, T2, I1, I2> ReadParams<(I1, I2)> for (T1, T2)
where
    T1: ReadParams<I1>,
    T2: ReadParams<I2>,
{
    fn read(params: &Parameters, index: (I1, I2)) -> Self {
        (T1::read(params, index.0), T2::read(params, index.1))
    }
}

impl<T1, T2, T3, I1, I2, I3> ReadParams<(I1, I2, I3)> for (T1, T2, T3)
where
    T1: ReadParams<I1>,
    T2: ReadParams<I2>,
    T3: ReadParams<I3>,
{
    fn read(params: &Parameters, index: (I1, I2, I3)) -> Self {
        (
            T1::read(params, index.0),
            T2::read(params, index.1),
            T3::read(params, index.2),
        )
    }
}

impl<T1, T2, T3, T4, I1, I2, I3, I4> ReadParams<(I1, I2, I3, I4)> for (T1, T2, T3, T4)
where
    T1: ReadParams<I1>,
    T2: ReadParams<I2>,
    T3: ReadParams<I3>,
    T4: ReadParams<I4>,
{
    fn read(params: &Parameters, index: (I1, I2, I3, I4)) -> Self {
        (
            T1::read(params, index.0),
            T2::read(params, index.1),
            T3::read(params, index.2),
            T4::read(params, index.3),
        )
    }
}
