use std::panic;

use bevy_ecs::world::World;
use tracing::Level;
use tracing_subscriber::{
    fmt::{format::DefaultFields, FormattedFields},
    prelude::*,
    EnvFilter, Registry,
};

pub fn setup(world: &mut World) {
    let old_handler = panic::take_hook();
    panic::set_hook(Box::new(move |infos| {
        println!("{}", tracing_error::SpanTrace::capture());
        old_handler(infos);
    }));

    let default_filter = { format!("{},{}", Level::INFO, "") };
    let filter_layer = EnvFilter::try_from_default_env()
        .or_else(|_| EnvFilter::try_new(&default_filter))
        .unwrap();
    let subscriber = Registry::default().with(filter_layer);

    let subscriber = subscriber.with(tracing_error::ErrorLayer::default());

    let chrome_layer = {
        let mut layer = tracing_chrome::ChromeLayerBuilder::new();
        if let Ok(path) = std::env::var("TRACE_CHROME") {
            layer = layer.file(path);
        }
        let (chrome_layer, guard) = layer
            .name_fn(Box::new(|event_or_span| match event_or_span {
                tracing_chrome::EventOrSpan::Event(event) => event.metadata().name().into(),
                tracing_chrome::EventOrSpan::Span(span) => {
                    if let Some(fields) = span.extensions().get::<FormattedFields<DefaultFields>>()
                    {
                        format!("{}: {}", span.metadata().name(), fields.fields.as_str())
                    } else {
                        span.metadata().name().into()
                    }
                }
            }))
            .build();
        world.insert_non_send_resource(guard);
        chrome_layer
    };
    let subscriber = subscriber.with(chrome_layer);

    tracing::subscriber::set_global_default(subscriber).unwrap();
}
